#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <array.h>

struct ARRAY_T
{ 
  size_t  size;
  void    *data[BLOCKSZ];
  ARRAY   next;
};

/**
 * "private" functions
 */
void         pri_destroy_node(ARRAY this, int index);
static ARRAY pri_array_locate(ARRAY this, int index, int *index_ptr);
static ARRAY pri_array_find( ARRAY this, int index, int *index_ptr );

ARRAY
new_array()
{
  ARRAY this;
  this       = calloc(sizeof(*this), 1);
  this->size = 0;
  this->next = NULL;
  return this;
}

size_t
array_get_size(ARRAY this)
{
  return this->size;
}

void
array_put(ARRAY this, int index, void *data)
{
  ARRAY  array;
  int    index_ptr;

  array = pri_array_locate( this, index, &index_ptr);
  array->data[index_ptr] = data;
  this->size++;
}

void *
array_get(ARRAY this, int index)
{
  ARRAY this_ptr;
  int   index_ptr;

  if(!this) return NULL;

  this_ptr = pri_array_locate(this, index, &index_ptr);
  if(this_ptr)
    return this_ptr->data[index_ptr];
  else
    return NULL;
}

ARRAY
array_destroy(ARRAY this)
{
  size_t x;
  size_t len;
  ARRAY prev, curr;

  if(this == NULL)
    return NULL;

  len = array_get_size(this);

  for(x = 0; x < len; x++){
    pri_destroy_node(this, x);
  }
  for(prev=NULL, curr=this; curr!=NULL; prev=curr, curr=curr->next){
    if(curr == this){
      this = curr->next;
      prev = this;
    } else {
      prev->next = curr->next;
    }
    free(curr);
    curr = prev;
    if(curr == NULL)
      break;
  }
  free(this);
  return NULL;
}

static ARRAY
pri_array_locate(ARRAY this, int index, int *index_ptr) {
  ARRAY array = this;
  *index_ptr  = index; 

  while(*index_ptr >= BLOCKSZ){
    if(array->next == NULL){
      array->next = calloc(sizeof(*array), 1);
    }
    array = array->next;
    *index_ptr -= BLOCKSZ;
  }
  return array;
}

static ARRAY
pri_array_find(ARRAY this, int index, int *index_ptr) {
  ARRAY array = this;
  *index_ptr  = index;

  while(*index_ptr >= BLOCKSZ){
    array = array->next;
    *index_ptr -= BLOCKSZ;
  }
  return array;
}

void 
pri_destroy_node(ARRAY this, int index) {
  ARRAY this_ptr;
  int   index_ptr;

  if(!this) return;

  this_ptr = pri_array_find(this, index, &index_ptr);
  if(this_ptr && index_ptr==BLOCKSZ){
    free(this_ptr->next);
  }
  return;
}

#if 0
int
pri_array_test()
{
  int   x;
  ARRAY A = new_array();

  for(x = 0; x < 100; x++){
    char tmp[32];
    snprintf(tmp, sizeof tmp, "haha=>%d", x);
    array_put(A, array_get_size(A), tmp);
  }
  for(x = 0; x < array_get_size(A); x++)
    printf("%s\n", array_get(A, x));
  A = array_destroy(A);

  return 0;
}

int
main()
{
  int x;
  for( x = 0; x < 5000; x++ )
    pri_array_test(); 
  return 0;
}
#endif
